/**
 * Remote control program robot.
 * 
 *      1 VDD        14 VSS
 * TX   2 RA5        13 RA0 DAT
 * IRin 3 RA4        12 RA1 CLK
 * MCLR 4 RA3        11 RA2 BIN2
 * SW   5 RC5        10 RC0 BIN1
 * SCL  6 RC4         9 RC1 AIN2
 * SDA  7 RC3         8 RC2 AIN1
 *  
 * PIC16F18326 MPLAB X IDE v5.40 with XC8 Ver2.30
 * (c)2020 Futoshi Sumikawa. 2020/09/29
 */

/**
    MCC

    System Module
      Current System clock 4 MHz

    TMR1
      Enable Timer
      -Timer Clock
        Clock Source FOSC/4
        Prescaler 1:2
        Enable Synchronization
      -Timer Period
        Timer Period 131.072 ms

    MEMORY
      Add DataEE Routines

    EUSART
      Mode asynchronous
      Enable EUSART
      Enable Transmit
      Baud Rate: 9600
      Transmission Bits: 8-bit
      Reception Bits: 8-bit
      Data Polarity Non-Inverted

    MSSP1
      -Hardware Settings
        Serial Protocol: I2C
        Mode: Master

    TMR0
      -Hardware Settings
        Enable Timer
          Clock prescaler 1:32768
          Postscaler 1:1
          Timer mode: 16-bit
          Clock Source: FOSC/4
          Enable Synchronization
      -Timer Period
        Requested Period: 100ms
        Actual Period: 98.304ms 
 
    TMR2
      -Hardware Settings
        Enable Timer
          Clock prescaler 1:10
          Postscaler 1:4
      -Timer Period
        Requested Period: 10ms
        Actual Period: 10ms

    PWM5
      -Hardware Settings
        Enable PWM5
        Select a Timer: Timer2

    PWM6
      -Hardware Settings
        Enable PWM6
        Select a Timer: Timer2
 
 */

/*
---------------------------------------------------------
DataEE Memory (0x7000)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - -
00 Initialize flag
... Reserve
10 Balance Forward
11 Balance Back
12 Balance Right
13 Balance Left
14 Unit Forward
15 Unit Back
16 Unit Right
17 Unit Left
18 Unit Wait
... Reserve
32 .. 255 IR Frame 01-15
---------------------------------------------------------
Flash Memory (0x3000)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - -
0000-00FF Programming #00-#07  
0100-01FF Programming #08-#15
0200-02FF Programming #16-#23
0300-03FF Programming #24-#31
0400-04FF Learning file 00
0500-05FF Learning file 01
0600-06FF Learning file 02
0700-07FF Learning file 03
0800-08FF Learning file 04
0900-09FF Learning file 05
0A00-0AFF Learning file 06
0B00-0BFF Learning file 07
0C00-0CFF Learning file 08
0D00-0DFF Learning file 09
0E00-0EFF Learning file 10
0F00-0FFF Reserve
--------------------------------------------------------- 
 */

/**
  Generated Main Source File

  Company:
    Microchip Technology Inc.

  File Name:
    main.c

  Summary:
    This is the main file generated using PIC10 / PIC12 / PIC16 / PIC18 MCUs

  Description:
    This header file provides implementations for driver APIs for all modules selected in the GUI.
    Generation Information :
        Product Revision  :  PIC10 / PIC12 / PIC16 / PIC18 MCUs - 1.81.6
        Device            :  PIC16F18326
        Driver Version    :  2.00
*/

/*
    (c) 2018 Microchip Technology Inc. and its subsidiaries. 
    
    Subject to your compliance with these terms, you may use Microchip software and any 
    derivatives exclusively with Microchip products. It is your responsibility to comply with third party 
    license terms applicable to your use of third party software (including open source software) that 
    may accompany Microchip software.
    
    THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS". NO WARRANTIES, WHETHER 
    EXPRESS, IMPLIED OR STATUTORY, APPLY TO THIS SOFTWARE, INCLUDING ANY 
    IMPLIED WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS 
    FOR A PARTICULAR PURPOSE.
    
    IN NO EVENT WILL MICROCHIP BE LIABLE FOR ANY INDIRECT, SPECIAL, PUNITIVE, 
    INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, COST OR EXPENSE OF ANY KIND 
    WHATSOEVER RELATED TO THE SOFTWARE, HOWEVER CAUSED, EVEN IF MICROCHIP 
    HAS BEEN ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE FORESEEABLE. TO 
    THE FULLEST EXTENT ALLOWED BY LAW, MICROCHIP'S TOTAL LIABILITY ON ALL 
    CLAIMS IN ANY WAY RELATED TO THIS SOFTWARE WILL NOT EXCEED THE AMOUNT 
    OF FEES, IF ANY, THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR THIS 
    SOFTWARE.
*/

#include "mcc_generated_files/mcc.h"
#include "irrxsync.h"
#include "st7032.h"
#include "drv8835.h"

#include "serial.h"

//#define ENABLE_EUSART
//#define ENABLE_LEARNING
//#define DISPLAY_ACTION_TIME

#define DATAEE_ADDRESS      0x7000
#define FLASH_ADDRESS       0x3000
#define FLASH_BLOCKSIZE     32

#define DATAEE_ADDRESS_BALANCE_F    0x7010
#define DATAEE_ADDRESS_BALANCE_B    0x7011
#define DATAEE_ADDRESS_BALANCE_R    0x7012
#define DATAEE_ADDRESS_BALANCE_L    0x7013
#define DATAEE_ADDRESS_UNIT_F       0x7014
#define DATAEE_ADDRESS_UNIT_B       0x7015
#define DATAEE_ADDRESS_UNIT_R       0x7016
#define DATAEE_ADDRESS_UNIT_L       0x7017
#define DATAEE_ADDRESS_UNIT_W       0x7018


#define SW_ON               0
#define SW_OFF              1


#define KEY_NONE            0
#define KEY_UP              1
#define KEY_RIGHT           2
#define KEY_DOWN            3
#define KEY_LEFT            4
#define KEY_ENTER           5
#define KEY_BACK            6

#define MODE_NONE           0
#define MODE_DRIVING        1
#define MODE_PROGRAMMING    2
#define MODE_RUN            3
#define MODE_LEARNING       4
#define MODE_PLAYBACK       5
                            
#define MOTION_NONE         0
#define MOTION_FORWARD      1
#define MOTION_BACK         2
#define MOTION_RIGHT        3
#define MOTION_LEFT         4
#define MOTION_WAIT         5

#define STATE_STOP          0
#define STATE_ACTION        1
#define STATE_PAUSE         2

#define PROGRAMMING_SELECT  0
#define PROGRAMMING_EDIT    1

#define MEMORY_FILE_OFFSET_INDEX 4

// Break time for playback 300ms
#define BREAK_TIME_FOR_PLAYBACK 3

#define DUTY_VALUE_FORWARD  799
#define DUTY_VALUE_BACK     799
#define DUTY_VALUE_RIGHT    449
#define DUTY_VALUE_LEFT     449

#define UNIT_F 10
#define UNIT_B 10
#define UNIT_R 4
#define UNIT_L 4
#define UNIT_W 10

const uint8_t CHAR_CUROSR[7] = {
  0b11111,
  0b11111,
  0b11111,
  0b11111,
  0b11111,
  0b11111,
  0b11111
};

const uint8_t CHAR_UP[7] = {
  0b00100,
  0b01110,
  0b10101,
  0b00100,
  0b00100,
  0b00100,
  0b00100
};

const uint8_t CHAR_DOWN[7] = {
  0b00100,
  0b00100,
  0b00100,
  0b00100,
  0b10101,
  0b01110,
  0b00100
};

const uint8_t CHAR_PAUSE[7] = {
  0b01010,
  0b01010,
  0b01010,
  0b01010,
  0b01010,
  0b01010,
  0b01010
};

const uint8_t CHAR_LEFT_BRACKET[7] = {
  0b00010,
  0b00100,
  0b00100,
  0b00100,
  0b00100,
  0b00100,
  0b00010
};

const uint8_t CHAR_RIGHT_BRACKET[7] = {
  0b01000,
  0b00100,
  0b00100,
  0b00100,
  0b00100,
  0b00100,
  0b01000
};

const uint8_t CHAR_STOP[7] = {
  0b00000,
  0b00000,
  0b11111,
  0b11111,
  0b11111,
  0b11111,
  0b11111
};

uint8_t _learningMotion = MOTION_NONE;
uint8_t _actionTime = 0;
uint8_t _fileIndex = 0;

uint8_t _playbackMotion = MOTION_NONE;

#ifdef ENABLE_LEARNING
uint8_t _enableLearning = true;
#else
uint8_t _enableLearning = false;
#endif

uint8_t _setup = 0;

// mode : driving / learning / playback / programming / run
uint8_t _mode = MODE_DRIVING;
// step : 1 - 3
uint8_t _step = 1;
// state : stop / action / pause
uint8_t _state = STATE_STOP;
// fileNumber : 1 - 10
uint8_t _fileNumber = 1;
// programNumber : 1 - 30
uint8_t _programNumber = 1;

// motion : none / forward / back / right / left / wait
uint8_t _motion = MOTION_NONE;

// pad : 0 - 6 (NURDLEB)
uint8_t _pad = 0;


// IR
uint8_t _isReceiveComplete = 0;
uint8_t irFrameData[32];
uint8_t irFrameDataPrevious[32];


// program
uint8_t _fileData[256];
uint8_t _programData[32];

uint8_t _cursorX = 0;
uint8_t _cursorY = 0;
uint8_t _edit = PROGRAMMING_SELECT;

int8_t _balanceF = 0;
int8_t _balanceB = 0;
int8_t _balanceR = 0;
int8_t _balanceL = 0;

uint8_t _unitF = 0;
uint8_t _unitB = 0;
uint8_t _unitR = 0;
uint8_t _unitL = 0;
uint8_t _unitW = 0;

uint8_t _breakTimeForPlayback = 0;

uint8_t _100ms = 0;

// ----------------------------------------
// MEMORY
// ----------------------------------------

uint16_t getFlashBlockAddress(uint8_t index) {
    uint16_t address = FLASH_ADDRESS + FLASH_BLOCKSIZE * index;
    return address;
}

uint16_t getFlashAddress(uint8_t index) {
    uint16_t address = FLASH_ADDRESS + 256 * index;
    return address;
}

void saveData(uint8_t index) {
    uint16_t address = getFlashAddress(index);
   
    uint16_t flashWordArray[FLASH_BLOCKSIZE];

    for (uint8_t n = 0; n < 8; n++) {
        uint8_t offset = n * FLASH_BLOCKSIZE;
        for (uint8_t i = 0; i < FLASH_BLOCKSIZE; i++) {
            flashWordArray[i] = _fileData[i + offset];
        }
        // write to Flash memory block
        FLASH_WriteBlock((uint16_t)(address + offset), (uint16_t*)flashWordArray);
    }
}

void loadData(uint8_t index) {
    uint16_t address = getFlashAddress(index);
    for (uint16_t i = 0; i <= 255; i++) {
        _fileData[i] = (uint8_t)FLASH_ReadWord(address + i);
    }
}

void saveProgram(uint8_t index) {
    uint16_t address = getFlashBlockAddress(index);
    
    uint16_t flashWordArray[FLASH_BLOCKSIZE];
    for (uint8_t i = 0; i < FLASH_BLOCKSIZE; i++) {
        flashWordArray[i] = _programData[i];
    }    
    
    // write to Flash memory block
    FLASH_WriteBlock((uint16_t)(address), (uint16_t*)flashWordArray);
}

void loadProgram(uint8_t index) {
    uint16_t address = getFlashBlockAddress(index);
    for (uint8_t i = 0; i < FLASH_BLOCKSIZE; i++) {
        uint8_t data = (uint8_t)FLASH_ReadWord(address + i);
        if (data == 0xff) {
            data = 0;
        }
        _programData[i] = data;
    }
}

// ----------------------------------------
// IR
// ----------------------------------------

uint8_t callbackIrInGetValue() {
    return IRin_GetValue();
}

void callbackResetTimer() {
    TMR1_WriteTimer(0);
}

uint16_t readTimer(void) {
    uint16_t readVal;
    uint8_t readValHigh;
    uint8_t readValLow;
	
    uint8_t preReadValLow = TMR1L;
    readValHigh = TMR1H;
    readValLow = TMR1L;
    if (readValLow < preReadValLow) {
        readValHigh = TMR1H;
    }
    readVal = ((uint16_t)readValHigh << 8) | readValLow;
    return readVal;
}

uint16_t callbackReadTimer() {
    return readTimer();
}


void serialSendFrameData() {
#ifdef ENABLE_EUSART
    for (uint8_t i = 0; i < 32; i++) {
        serialPutHexData(irFrameData[i]);
    }
    
    serialPutCR();
    serialPutLF();
#endif
}

void serialSendFrameDataDATAEE(uint8_t index) {
    uint16_t address = (uint16_t)(DATAEE_ADDRESS + (index << 5));
    
    for (uint8_t i = 0; i < 32; i++) {
        serialPutHexData(DATAEE_ReadByte(address  + i));
    }
    
    serialPutCR();
    serialPutLF();
}

void copyPreviousFrameData() {
    for (uint8_t i = 0; i < 32; i++) {
        irFrameDataPrevious[i] = irFrameData[i];
    }
}

uint8_t comparePreviousFrameCodeData() {

    for (uint8_t i = 0; i < 16; i++) {
        if (irFrameData[16 + i] != irFrameDataPrevious[16 + i]) {
            return 0;
        }
    }
    return 1;
}

uint8_t compareFrameCodeData(uint8_t index) {
    uint16_t address = (uint16_t)(DATAEE_ADDRESS + (index << 5));

    for (uint8_t i = 0; i < 16; i++) {
        if (irFrameData[16 + i] != DATAEE_ReadByte(address + 16 + i)) {
            return 0;
        }
    }
    return 1;
}

void saveFrameData(uint8_t index) {
    uint16_t address = (uint16_t)(DATAEE_ADDRESS + (index << 5));

    for (uint8_t i = 0; i < 32; i++) {
        DATAEE_WriteByte(address + i, irFrameData[i]);
    }
}


uint8_t inputKey() {
    uint8_t result = 0;
    for (uint8_t i = 1; i < 7; i++) {
       if (compareFrameCodeData(i) == 1) {
           result = i;
           break;
       }
    }
    return result;
}

// ----------------------------------------
// SETUP
// ----------------------------------------

void setupController() {
    if (compareFrameCodeData(_setup - 1) == 0) {
        if (comparePreviousFrameCodeData()) {
            saveFrameData(_setup);
            if (_setup < 6) {
                _setup++;
            } else {
                _setup = 0;
            }
        } else {
            copyPreviousFrameData();
        }
    }
}

void setupBalance() {
    uint8_t key = inputKey();
    if (_pad != key) {
        _pad = key;
        
        if (key == KEY_RIGHT) {
            if (_setup < 10) {
                _setup++;
            } else {
                _setup = 7;
            }
        } else if (key == KEY_LEFT) {
            if (_setup > 7) {
                _setup--;
            } else {
                _setup = 10;
            }
        } else if (key == KEY_UP) {
            if (_setup == 7) {
                if (_balanceF < 99) {
                    _balanceF++;
                }
                drvSetBalanceForward(_balanceF);
                DATAEE_WriteByte(DATAEE_ADDRESS_BALANCE_F, (uint8_t)_balanceF);
            } else if (_setup == 8) {
                if (_balanceB < 99) {
                    _balanceB++;
                }
                drvSetBalanceBack(_balanceB);
                DATAEE_WriteByte(DATAEE_ADDRESS_BALANCE_B, (uint8_t)_balanceB);
            } else if (_setup == 9) {
                if (_balanceR < 99) {
                    _balanceR++;
                }
                drvSetBalanceRight(_balanceR);
                DATAEE_WriteByte(DATAEE_ADDRESS_BALANCE_R, (uint8_t)_balanceR);
            } else if (_setup == 10) {
                if (_balanceL < 99) {
                    _balanceL++;
                }
                drvSetBalanceLeft(_balanceL);
                DATAEE_WriteByte(DATAEE_ADDRESS_BALANCE_L, (uint8_t)_balanceL);
            }
        } else if (key == KEY_DOWN) {
            if (_setup == 7) {
                if (_balanceF > -99) {
                    _balanceF--;
                }
                drvSetBalanceForward(_balanceF);
                DATAEE_WriteByte(DATAEE_ADDRESS_BALANCE_F, (uint8_t)_balanceF);
            } else if (_setup == 8) {
                if (_balanceB > -99) {
                    _balanceB--;
                }
                drvSetBalanceBack(_balanceB);
                DATAEE_WriteByte(DATAEE_ADDRESS_BALANCE_B, (uint8_t)_balanceB);
            } else if (_setup == 9) {
                if (_balanceR > -99) {
                    _balanceR--;
                }
                drvSetBalanceRight(_balanceR);
                DATAEE_WriteByte(DATAEE_ADDRESS_BALANCE_R, (uint8_t)_balanceR);
            } else if (_setup == 10) {
                if (_balanceL > -99) {
                    _balanceL--;
                }
                drvSetBalanceLeft(_balanceL);
                DATAEE_WriteByte(DATAEE_ADDRESS_BALANCE_L, (uint8_t)_balanceL);
            }            
        } else if (key == KEY_ENTER || key == KEY_BACK) {
            _setup = 0;
        }
    }
}

void setupUnit() {
    uint8_t key = inputKey();
    if (_pad != key) {
        _pad = key;
        
        if (key == KEY_RIGHT) {
            if (_setup < 15) {
                _setup++;
            } else {
                _setup = 11;
            }
        } else if (key == KEY_LEFT) {
            if (_setup > 11) {
                _setup--;
            } else {
                _setup = 15;
            }
        } else if (key == KEY_UP) {
            if (_setup == 11) {
                if (_unitF < 32) {
                    _unitF++;
                    DATAEE_WriteByte(DATAEE_ADDRESS_UNIT_F, _unitF);
                }
            } else if (_setup == 12) {
                if (_unitB < 32) {
                    _unitB++;
                    DATAEE_WriteByte(DATAEE_ADDRESS_UNIT_B, _unitB);
                }
            } else if (_setup == 13) {
                if (_unitR < 32) {
                    _unitR++;
                    DATAEE_WriteByte(DATAEE_ADDRESS_UNIT_R, _unitR);
                }
            } else if (_setup == 14) {
                if (_unitL < 32) {
                    _unitL++;
                    DATAEE_WriteByte(DATAEE_ADDRESS_UNIT_L, _unitL);
                }
            } else if (_setup == 15) {
                if (_unitW < 32) {
                    _unitW++;
                    DATAEE_WriteByte(DATAEE_ADDRESS_UNIT_W, _unitW);
                }
            }
        } else if (key == KEY_DOWN) {
            if (_setup == 11) {
                if (_unitF > 1) {
                    _unitF--;
                    DATAEE_WriteByte(DATAEE_ADDRESS_UNIT_F, _unitF);
                }
            } else if (_setup == 12) {
                if (_unitB > 1) {
                    _unitB--;
                    DATAEE_WriteByte(DATAEE_ADDRESS_UNIT_B, _unitB);
                }
            } else if (_setup == 13) {
                if (_unitR > 1) {
                    _unitR--;
                    DATAEE_WriteByte(DATAEE_ADDRESS_UNIT_R, _unitR);
                }
            } else if (_setup == 14) {
                if (_unitL > 1) {
                    _unitL--;
                    DATAEE_WriteByte(DATAEE_ADDRESS_UNIT_L, _unitL);
                }
            } else if (_setup == 15) {
                if (_unitW > 1) {
                    _unitW--;
                    DATAEE_WriteByte(DATAEE_ADDRESS_UNIT_W, _unitW);
                }
            }
        } else if (key == KEY_ENTER || key == KEY_BACK) {
            _setup = 0;
        }
    }
}

// ----------------------------------------
// ROBOT
// ----------------------------------------

uint8_t motionKey(uint8_t key) {
    if (key == KEY_UP) {
        return MOTION_FORWARD;
    } else if (key == KEY_RIGHT) {
        return MOTION_RIGHT;
    } else if (key == KEY_DOWN) {
        return MOTION_BACK;
    } else if (key == KEY_LEFT) {
        return MOTION_LEFT;
    }
    return MOTION_WAIT;
}

void move(uint8_t motion) {
    if (motion == MOTION_WAIT) {
        drvWait();
    } else if (motion == MOTION_FORWARD) {
        drvForward(DUTY_VALUE_FORWARD);
    } else if (motion == MOTION_BACK) {
        drvBack(DUTY_VALUE_BACK);
    } else if (motion == MOTION_RIGHT) {
        drvRight(DUTY_VALUE_RIGHT);
    } else if (motion == MOTION_LEFT) {
        drvLeft(DUTY_VALUE_LEFT);
    }
}

void loadInitialProgram() {
    uint8_t pg[] = {
        0x12, 0x21, 0x34, 0x46,
        0x32, 0x53, 0x14, 0x25,
        0x11, 0x00
    };
    for (uint8_t i = 0; i < 10; i++) {
        _programData[i] = pg[i];
    }
}

void clearProgram() {
    for (uint8_t i = 0; i < 16; i++) {
        _programData[i] = 0;
    }
}

void deployProgram() {
    uint8_t ix = 0;
    for (uint8_t i = 0; i < 16; i++) {
        uint8_t d = _programData[i];
        if (d != 0) {
            uint8_t motion = (d & 0xf0) >> 4;
            uint8_t unit = 10;
            if (motion == MOTION_FORWARD) {
                unit = _unitF;
            } else if (motion == MOTION_BACK) {
                unit = _unitB;
            } else if (motion == MOTION_RIGHT) {
                unit = _unitR;
            } else if (motion == MOTION_LEFT) {
                unit = _unitL;
            } else if (motion == MOTION_WAIT) {
                unit = _unitW;
            }
            uint8_t m = (uint8_t)((motion << 5) + unit);
            uint8_t n = d & 0x0f;
            while (n > 0) {
                _fileData[ix++] = m;
                n--;
            }
        } else {
            _fileData[ix] = 0;
            break;
        }
    }
}

void startLearning() {
    _state = STATE_ACTION;
    _actionTime = 0;
    _fileIndex = 0;
    for (uint8_t i = 0; i <= 254; i++) {
        _fileData[i] = 0;
    }
    _fileData[255] = 0;
}

void finishLearning() {
    _state = STATE_STOP;
    // TODO
    saveData(_fileNumber + MEMORY_FILE_OFFSET_INDEX);
}

void startPlayback() {
    // TODO
    loadData(_fileNumber + MEMORY_FILE_OFFSET_INDEX);
    _state = STATE_ACTION;
    _actionTime = 0;
    _fileIndex = 0;
    _playbackMotion = MOTION_NONE;
    _breakTimeForPlayback = 0;
}

void finishPlayback() {
    _state = STATE_STOP;

}

void pausePlayback() {
    if (_state == STATE_PAUSE) {
        _state = STATE_ACTION;
    } else {
        _state = STATE_PAUSE;
    }
}


void startRun() {
    loadProgram(_programNumber);
    
    deployProgram();
    _state = STATE_ACTION;
    _actionTime = 0;
    _fileIndex = 0;
    _playbackMotion = MOTION_NONE;
    _breakTimeForPlayback = 0;
}

void finishRun() {
    _state = STATE_STOP;

}

void pauseRun() {
    if (_state == STATE_PAUSE) {
        _state = STATE_ACTION;
    } else {
        _state = STATE_PAUSE;
    }
}

void setViewState() {
    uint8_t key = inputKey();
    if (_pad != key) {
        _pad = key;
        
        if (_step == 1) {
            // Select mode.
            if (key == KEY_RIGHT) {
                if (_enableLearning) {
                    if (_mode == MODE_PLAYBACK) {
                        _mode = MODE_DRIVING;
                    } else {
                        _mode++;
                    }
                } else {
                    if (_mode == MODE_RUN) {
                        _mode = MODE_DRIVING;
                    } else {
                        _mode++;
                    }
                }
            } else if (key == KEY_LEFT) {
                if (_enableLearning) {
                    if (_mode == MODE_DRIVING) {
                        _mode = MODE_PLAYBACK;
                    } else {
                        _mode--;
                    }
                } else {
                    if (_mode == MODE_DRIVING) {
                        _mode = MODE_RUN;
                    } else {
                        _mode--;
                    }
                }
            } else if (key == KEY_ENTER) {
                _step = 2;
            }

        } else if (_step == 2) {
            if (_mode == MODE_LEARNING || _mode == MODE_PLAYBACK) {
                if (key == KEY_UP) {
                    _fileNumber++;
                    if (_fileNumber > 10) {
                        _fileNumber = 1;
                    }
                } else if (key == KEY_DOWN) {
                    _fileNumber--;
                    if (_fileNumber < 1) {
                        _fileNumber = 10;
                    }
                }
            } else if (_mode == MODE_PROGRAMMING || _mode == MODE_RUN) {
                if (key == KEY_UP) {
                    _programNumber++;
                    if (_programNumber > 30) {
                        _programNumber = 1;
                    }
                } else if (key == KEY_DOWN) {
                    _programNumber--;
                    if (_programNumber < 1) {
                        _programNumber = 30;
                    }
                }
            }
            
            if (key == KEY_ENTER) {
                if (_mode == MODE_LEARNING) {
                    _step = 3;
                    startLearning();
                } else if (_mode == MODE_PLAYBACK) {
                    _step = 3;
                    startPlayback();
                } else if (_mode == MODE_PROGRAMMING) {
                    _step = 3;
                    loadProgram(_programNumber);
                } else if (_mode == MODE_RUN) {
                    _step = 3;
                    startRun();
                }
                
            } else if (key == KEY_BACK) {
                _step = 1;
            }

        } else if (_step == 3) {
            if (_mode == MODE_LEARNING) {
                if (key == KEY_BACK) {
                    _step = 2;
                    finishLearning();
                } 
            } else if (_mode == MODE_PLAYBACK) {
                if (key == KEY_BACK) {
                    _step = 2;
                    finishPlayback();
                } else if (key == KEY_ENTER) {
                    pausePlayback();
                } 

            } else if (_mode == MODE_PROGRAMMING) {

                if (_edit == PROGRAMMING_EDIT) {
                    if (_cursorY == 0) {
                        // motion

                        uint8_t d = _programData[_cursorX];
                        uint8_t m = (d & 0xf0) >> 4;
                        
                        if (key == KEY_UP) {
                            if (m == MOTION_FORWARD) {
                                m = MOTION_WAIT;
                            } else {
                                m = MOTION_FORWARD;
                            }
                        } else if (key == KEY_RIGHT) {
                            if (m == MOTION_RIGHT) {
                                m = MOTION_WAIT;
                            } else {
                                m = MOTION_RIGHT;
                            }
                        } else if (key == KEY_DOWN) {
                            if (m == MOTION_BACK) {
                                m = MOTION_WAIT;
                            } else {
                                m = MOTION_BACK;
                            }
                        } else if (key == KEY_LEFT) {
                            if (m == MOTION_LEFT) {
                                m = MOTION_WAIT;
                            } else {
                                m = MOTION_LEFT;
                            }
                        }
                        
                        _programData[_cursorX] = (uint8_t)((d & 0x0f) | (m << 4));
                        
                    } else {
                        // time
                        
                        uint8_t d = _programData[_cursorX];
                        uint8_t t = d & 0x0f;
                        
                        if (key == KEY_UP || key == KEY_RIGHT) {
                            if (t < 15) {
                                t++;
                            }
                        } else if (key == KEY_DOWN || key == KEY_LEFT) {
                            if (t > 0) {
                                t--;
                            }
                        }
                        
                        _programData[_cursorX] = (d & 0xf0) + t;
                        
                    }

                    if (key == KEY_ENTER || key == KEY_BACK) {
                        _edit = PROGRAMMING_SELECT;
                    } 
                    
                } else {
                    // cursor
                    
                    if (key == KEY_UP || key == KEY_DOWN) {
                        if (_cursorY == 0) {
                            _cursorY = 1;
                        } else {
                            _cursorY = 0;
                        }
                    } else if (key == KEY_RIGHT) {
                        if (_cursorX == 15) {
                            _cursorX = 0;
                        } else {
                            _cursorX++;
                        }
                    } else if (key == KEY_LEFT) {
                        if (_cursorX == 0) {
                            _cursorX = 15;
                        } else {
                            _cursorX--;
                        }
                    }
                    
                    if (key == KEY_BACK) {
                        _step = 2;
                        saveProgram(_programNumber);
                    } else if (key == KEY_ENTER) {
                        _edit = PROGRAMMING_EDIT;
                    } 
                }

            } else if (_mode == MODE_RUN) {
                if (key == KEY_BACK) {
                    _step = 2;
                    finishRun();
                } else if (key == KEY_ENTER) {
                    pauseRun();
                } 

            }
            
        }
    }            
}

void readLearningData(void) {
    uint8_t data = _fileData[_fileIndex];
    _actionTime = data & 0x1f;
    _playbackMotion = data >> 5;
    _fileIndex++;
}

uint8_t _maximumActionTime = 0;
uint8_t _invalideteDisplay = 1;


void cycle100ms(void) {
    
    // Call every 100ms or more.
    
    _100ms++;
    _invalideteDisplay = 1;
    
    // Pad
    
    if (_isReceiveComplete) {
        _isReceiveComplete = 0;
    } else {
        _pad = KEY_NONE;
    }

    if (_setup != 0) {
        return;
    }
    
    if (_mode == MODE_DRIVING) {
        
        // Driving
        
        if (_step == 2) {
            _motion = motionKey(_pad);
            move(_motion);
        }
        
    } else if (_mode == MODE_LEARNING) {
        
        // Learning
        
        if (_step == 3 && _state == STATE_ACTION) {
            _motion = motionKey(_pad);
            move(_motion);
            uint8_t learningMotion = _motion;

            _actionTime++;

            if (_actionTime >= 31 || _learningMotion != learningMotion) {
                uint8_t data = (uint8_t)(_actionTime | (learningMotion << 5));
                _fileData[_fileIndex++] = data;
                if (_fileIndex == 0) {
                    // finish learning
                    _state = STATE_STOP;
                }
                _actionTime = 0;
            }
            _learningMotion = learningMotion;
        }

    } else if (_mode == MODE_PLAYBACK || _mode == MODE_RUN) {
        
        // Playback

        if (_step == 3 && _state == STATE_ACTION) {
            if (_actionTime == 0) {
                if ((_playbackMotion == (_fileData[_fileIndex] >> 5 )) && _maximumActionTime) {
                    _breakTimeForPlayback = 0;
                }
                if (_breakTimeForPlayback) {
                    _breakTimeForPlayback = 0;
                    // Break time for playback
                    _actionTime = BREAK_TIME_FOR_PLAYBACK;
                    _playbackMotion = MOTION_WAIT;
                } else {
                    _breakTimeForPlayback = 1;
                    // Read learning data
                    uint8_t data = _fileData[_fileIndex];
                    _actionTime = data & 0x1f;
                    if (_actionTime == 32) {
                        _maximumActionTime = 1;
                    } else {
                        _maximumActionTime = 0;
                    }
                    _playbackMotion = data >> 5;
                    _fileIndex++;           
                }
            }
            _motion = _playbackMotion;
            move(_motion);
            if (_playbackMotion == MOTION_NONE || _fileIndex == 0) {
                finishPlayback();
            } else {
                _actionTime--;
            }
        } else {
            // Pause
            _motion = MOTION_WAIT;
            move(_motion);
        }        

    }
    
}

/*
                         Main application
 */
void main(void)
{
    // initialize the device
    SYSTEM_Initialize();

    TMR0_SetInterruptHandler(cycle100ms);
    
    // When using interrupts, you need to set the Global and Peripheral Interrupt Enable bits
    // Use the following macros to:

    // Enable the Global Interrupts
    INTERRUPT_GlobalInterruptEnable();

    // Enable the Peripheral Interrupts
    INTERRUPT_PeripheralInterruptEnable();

    // Disable the Global Interrupts
    //INTERRUPT_GlobalInterruptDisable();

    // Disable the Peripheral Interrupts
    //INTERRUPT_PeripheralInterruptDisable();

    // Switch.
    if (SW1_GetValue() == SW_ON) {
        __delay_ms(20);
        if (SW1_GetValue() == SW_ON) {
           _enableLearning = true;
        }
    }
    
    uint8_t eeInitializeFlag = DATAEE_ReadByte(DATAEE_ADDRESS);
    if (eeInitializeFlag == 0 || eeInitializeFlag == 0xff) {
        DATAEE_WriteByte(DATAEE_ADDRESS_BALANCE_F, 0);
        DATAEE_WriteByte(DATAEE_ADDRESS_BALANCE_B, 0);
        DATAEE_WriteByte(DATAEE_ADDRESS_BALANCE_R, 0);
        DATAEE_WriteByte(DATAEE_ADDRESS_BALANCE_L, 0);
        DATAEE_WriteByte(DATAEE_ADDRESS_UNIT_F, UNIT_F);
        DATAEE_WriteByte(DATAEE_ADDRESS_UNIT_B, UNIT_B);
        DATAEE_WriteByte(DATAEE_ADDRESS_UNIT_R, UNIT_R);
        DATAEE_WriteByte(DATAEE_ADDRESS_UNIT_L, UNIT_L);
        DATAEE_WriteByte(DATAEE_ADDRESS_UNIT_W, UNIT_W);
        
        loadInitialProgram();
        saveProgram(1);
        
        DATAEE_WriteByte(DATAEE_ADDRESS, 1);
    } else {
        _balanceF = (int8_t)DATAEE_ReadByte(DATAEE_ADDRESS_BALANCE_F);
        _balanceB = (int8_t)DATAEE_ReadByte(DATAEE_ADDRESS_BALANCE_B);
        _balanceR = (int8_t)DATAEE_ReadByte(DATAEE_ADDRESS_BALANCE_R);
        _balanceL = (int8_t)DATAEE_ReadByte(DATAEE_ADDRESS_BALANCE_L);
        _unitF = DATAEE_ReadByte(DATAEE_ADDRESS_UNIT_F);
        _unitB = DATAEE_ReadByte(DATAEE_ADDRESS_UNIT_B);
        _unitR = DATAEE_ReadByte(DATAEE_ADDRESS_UNIT_R);
        _unitL = DATAEE_ReadByte(DATAEE_ADDRESS_UNIT_L);
        _unitW = DATAEE_ReadByte(DATAEE_ADDRESS_UNIT_W);
    }
    drvSetBalanceForward(_balanceF);
    drvSetBalanceBack(_balanceB);
    drvSetBalanceRight(_balanceR);
    drvSetBalanceLeft(_balanceL);
    
    irrxBegin(irFrameData, callbackIrInGetValue, callbackResetTimer, callbackReadTimer);    

    _mode = MODE_DRIVING;
    
    
    lcdBegin();
    lcdSetContrast(48);
    
    lcdRegisterChar(1, (uint8_t*)CHAR_CUROSR);
    lcdRegisterChar(2, (uint8_t*)CHAR_UP);
    lcdRegisterChar(3, (uint8_t*)CHAR_DOWN);
    lcdRegisterChar(4, (uint8_t*)CHAR_PAUSE);
    lcdRegisterChar(5, (uint8_t*)CHAR_LEFT_BRACKET);
    lcdRegisterChar(6, (uint8_t*)CHAR_RIGHT_BRACKET);
    lcdRegisterChar(7, (uint8_t*)CHAR_STOP);

    clearProgram();
 
    while (1)
    {
        // Add your application code

        if (irrxIsReceiveComplete() == 1) {
            // Receive.
            
            _isReceiveComplete = 1;
            
            if (_setup != 0) {
                if (_setup < 7) {
                    // Controller setup
                    setupController();
                } else if (_setup >= 7 && _setup <= 10) {
                    // Motor balance
                    setupBalance();
                } else if (_setup >= 11 && _setup <= 15) {
                    setupUnit();
                }
            } else {
                setViewState();
            }

            // Send frame data.
            serialSendFrameData();

            _invalideteDisplay = 1;
        }
        
        // Switch.
        if (SW1_GetValue() == SW_ON) {
            __delay_ms(20);
            while (SW1_GetValue() == SW_ON) {
                // Wait for button up.
            }
            if (_setup == 0) {
                _setup = 1;
                _invalideteDisplay = 1;
            } else if (_setup < 7) {
                _setup = 7;
            } else if (_setup < 11) {
                _setup = 11;
            } else {
                _setup = 0;
            }
        }

        // Display.
        if (_invalideteDisplay == 1) {
            _invalideteDisplay = 0;

            if (_setup != 0) {
                // Display controller setup
                lcdSetCursor(0,0);
                if (_setup < 7) {
                    lcdPuts("Controller setup");
                } else if (_setup >= 7 && _setup <= 10) {
                    lcdPuts("Motor balance   ");
                } else if (_setup >= 11 && _setup <= 15) {
                    lcdPuts("Unit time       ");
                }
                lcdSetCursor(0,1);
                if (_setup == 1) {
                    // Press UP key
                    lcdPuts("Press UP key    ");
                } else if (_setup == 2) {
                    // Press RIGHT key
                    lcdPuts("Press RIGHT key ");
                } else if (_setup == 3) {
                    // Press DOWN key
                    lcdPuts("Press DOWN key  ");
                } else if (_setup == 4) {
                    // Press LEFT key
                    lcdPuts("Press LEFT key  ");
                } else if (_setup == 5) {
                    // Press ENTER key
                    lcdPuts("Press ENTER key ");
                } else if (_setup == 6) {
                    // Press BACK key
                    lcdPuts("Press BACK key  ");
                } else if (_setup >= 7 && _setup <= 10) {
                    // Motor balance
                    int8_t balance = 0;
                    if(_setup == 7) {
                        balance = _balanceF;
                        lcdPuts("Forward: ");
                    } else if (_setup == 8) {
                        balance = _balanceB;
                        lcdPuts("Back:    ");
                    } else if (_setup == 9) {
                        balance = _balanceR;
                        lcdPuts("Right:   ");
                    } else if (_setup == 10) {
                        balance = _balanceL;
                        lcdPuts("Left:    ");
                    }
                    uint8_t value = 0;
                    if (balance < 0) {
                        lcdPutc('-');
                        value = (uint8_t)(balance * (-1));
                    } else {
                        lcdPutc(' ');
                        value = (uint8_t)balance;
                    }
                    uint8_t ten = value / 10;
                    uint8_t one = value % 10;
                    lcdPutc('0' + ten);
                    lcdPutc('0' + one);
                    lcdPuts("    ");

                } else if (_setup >= 11 && _setup <= 15) {
                    // Unit time
                    uint8_t unit = 0;
                    if(_setup == 11) {
                        unit = _unitF;
                        lcdPuts("Forward: ");
                    } else if (_setup == 12) {
                        unit = _unitB;
                        lcdPuts("Back:    ");
                    } else if (_setup == 13) {
                        unit = _unitR;
                        lcdPuts("Right:   ");
                    } else if (_setup == 14) {
                        unit = _unitL;
                        lcdPuts("Left:    ");
                    } else if (_setup == 15) {
                        unit = _unitW;
                        lcdPuts("Wait:    ");
                    }
                    uint8_t ten = unit / 10;
                    uint8_t one = unit % 10;
                    if (ten != 0) {
                        lcdPutc('0' + ten);
                    } else {
                        lcdPutc(' ');
                    }
                    lcdPutc('0' + one);
                    lcdPuts("00ms ");
                }
            } else {
                // mode.
                if (_step == 3 && _mode == MODE_PROGRAMMING) {
                    // Display program
                    for (uint8_t i = 0; i < 16; i++) {
                        uint8_t d = _programData[i];
                        uint8_t m = d >> 4;
                        uint8_t n = d & 0x0f;

                        uint8_t cursorOn = 0;
                        if (_edit == PROGRAMMING_SELECT) {
                            if (_100ms & 0x02) {
                                cursorOn = 1;
                            }
                        } else {
                            if (_100ms & 0x04) {
                                cursorOn = 1;
                            }
                        }
                        
                        lcdSetCursor(i,0);
                        if (_cursorX == i && _cursorY == 0 && cursorOn) {
                            if (_edit == PROGRAMMING_SELECT) {
                                lcdPutc(0x01);
                            } else {
                                lcdPutc(' ');
                            }
                        } else {
                            if (m == MOTION_WAIT) {
                                lcdPutc('o');
                            } else if (m == MOTION_FORWARD) {
                                lcdPutc(0x02);
                            } else if (m == MOTION_BACK) {
                                lcdPutc(0x03);
                            } else if (m == MOTION_RIGHT) {
                                lcdPutc(0x7e);
                            } else if (m == MOTION_LEFT) {
                                lcdPutc(0x7f);
                            } else {
                                lcdPutc(' ');
                            }
                        }
                        lcdSetCursor(i,1);
                        if (_cursorX == i && _cursorY == 1 && cursorOn) {
                            if (_edit == PROGRAMMING_SELECT) {
                                lcdPutc(0x01);
                            } else {
                                lcdPutc(' ');
                            }
                        } else {
                            if (n < 10) {
                                lcdPutc('0' + n);
                            } else {
                                lcdPutc('A' + (n - 10));
                            }
                        }
                    }

                } else {
                    // Display mode
                    lcdSetCursor(0,0);
                    if (_mode == MODE_DRIVING) {
                        lcdPuts("Driving    ");
                    } else if (_mode == MODE_LEARNING) {
                        lcdPuts("Learning   ");
                    } else if (_mode == MODE_PLAYBACK) {
                        lcdPuts("Playback   ");
                    } else if (_mode == MODE_PROGRAMMING) {
                        lcdPuts("Programming");
                    } else if (_mode == MODE_RUN) {
                        lcdPuts("Run        ");
                    }
                }

                // number.
                if (_step == 3 && _mode == MODE_PROGRAMMING) {
                    // Show nothing
                } else if (_step == 2 || _step == 3) {
                    if (_mode == MODE_LEARNING || _mode == MODE_PLAYBACK) {
                        // Display file number
                        lcdSetCursor(9,0);
                        lcdPuts("file ");
                        if (_fileNumber == 10) {
                            lcdPuts("10");
                        } else {
                            lcdPutc('0');
                            lcdPutc('0' + _fileNumber);
                        }
                    } else if (_mode == MODE_PROGRAMMING || _mode == MODE_RUN) {
                        // Display program number
                        lcdPuts("  #");
                        uint8_t ten = _programNumber / 10;
                        uint8_t one = _programNumber % 10;
                        lcdPutc('0' + ten);
                        lcdPutc('0' + one);
                    } else {
                        lcdPuts("     ");
                    }
                } else {
                    lcdPuts("     ");
                }

                // button.
                if (_step == 1) {
                    // Display button
                    lcdSetCursor(0,1);
                    lcdPuts("  [");
                    lcdPutc(0x7f);
                    lcdPuts("][");
                    lcdPutc(0x7e);
                    lcdPuts("]/[ENTER]");
                } else if (_step == 2 && _mode != MODE_DRIVING) {
                    // Display button
                    lcdSetCursor(0,1);
                    lcdPuts("  [");
                    lcdPutc(0x02);
                    lcdPuts("][");
                    lcdPutc(0x03);
                    lcdPuts("]/[ENTER]");
                }                
                // motion.
                if ((_step == 2 && _mode == MODE_DRIVING) || (_step == 3 && _mode != MODE_PROGRAMMING)) {
                    // Display motion
                    lcdSetCursor(0,1);
                    if (_mode == MODE_DRIVING || _state == STATE_ACTION) {
                        lcdPutc(0x05);
                        if (_motion == MOTION_WAIT) {
                            lcdPutc('o');
                            lcdPutc(0x06);
                            lcdPuts("Wait   ");
                        } else if (_motion == MOTION_FORWARD) {
                            lcdPutc(0x02);
                            lcdPutc(0x06);
                            lcdPuts("Forward");
                        } else if (_motion == MOTION_BACK) {
                            lcdPutc(0x03);
                            lcdPutc(0x06);
                            lcdPuts("Back   ");
                        } else if (_motion == MOTION_RIGHT) {
                            lcdPutc(0x7e);
                            lcdPutc(0x06);
                            lcdPuts("Right  ");
                        } else if (_motion == MOTION_LEFT) {
                            lcdPutc(0x7f);
                            lcdPutc(0x06);
                            lcdPuts("Left   ");
                        }
                    } else if(_state == STATE_PAUSE) {
                        lcdPutc(0x05);
                        lcdPutc(0x04);
                        lcdPutc(0x06);
                        lcdPuts("Pause  ");
                    } else {
                        lcdPutc(0x05);
                        lcdPutc(0x07);
                        lcdPutc(0x06);
                        lcdPuts("End    ");
                    }
                }

                // action time.
        #ifdef DISPLAY_ACTION_TIME
                if (_step == 3 && _mode != MODE_PROGRAMMING) {
                    // Display action time
                    if (_actionTime == 0 || _state != STATE_ACTION) {
                        lcdPuts("      ");
                    } else {
                        uint8_t ten = _actionTime / 10;
                        uint8_t one = _actionTime - (ten * 10);
                        if (ten == 0) {
                            lcdPutc(' ');
                        } else {
                            lcdPutc('0' + ten);
                        }
                        lcdPutc('0' + one);
                        lcdPuts("00ms");
                    }
                }
                if (_step == 2 && _mode == MODE_DRIVING) {
                    // Clear action time
                    lcdPuts("      ");
                }
        #else
                // Clear action time
                lcdPuts("      ");        
        #endif
            }

        }
    }
}


/**
 End of File
*/